#include <hell_tables.hh>
#include <integration.hh>
#include <gsl/gsl_errno.h>

namespace HELL {
  
  //const Real HELL_N_INTERVALS = 10000;
  //const Real HELL_N_STEP = 0.0000998;
  
  GammaHellTable & GammaHellTable::read(std::istream & in)
  {
    Real tmp;
    in >> alphas_ >> nf_;
    //
    int npoles;
    in >> npoles;
    poles_.resize(npoles);
    for (int i=0; i<npoles; ++i)
      in >> poles_[i].pole;
    for (int i=0; i<npoles; ++i)
      in >> poles_[i].residue;
    //
    int npoints;
    in >> npoints >> step_;
    gamma_.resize(npoints);
    in >> c_;
    in >> tmp;  eith_  =   tmp;
    in >> tmp;  eith_ += I*tmp;
    //
    for (int i=0; i<npoints; ++i) {
      in >> tmp;  gamma_[i]  =   tmp;
      in >> tmp;  gamma_[i] += I*tmp;
    }
    return *this;
  }
  Complex GammaHellTable::getN(unsigned i) const {
    Real u = i*step_;
    Real t = u/(1-u);
    return c_ + t*eith_ -1.; // the -1 restores the small-N notation for which the poles are in N=0
  }



  /// HellTableP

  HellTableP & HellTableP::readTable(std::string in) {
    std::ifstream infile(in.c_str());
    if (!infile.is_open())
      std::cerr << "error: file " << in << " does not exist" << std::endl;
    readTable(infile);
    return *this;
  }
  HellTableP & HellTableP::readDotTable(std::string in) {
    std::ifstream infile(in.c_str());
    if (!infile.is_open())
      std::cerr << "error: file " << in << " does not exist" << std::endl;
    readDotTable(infile);
    return *this;
  }
  HellTableP & HellTableP::readNLOTable(std::string in) {
    std::ifstream infile(in.c_str());
    if (!infile.is_open())
      std::cerr << "error: file " << in << " does not exist" << std::endl;
    readNLOTable(infile);
    return *this;
  }

  unsigned HellTableP::getTableSize() const {
    unsigned size = g_.getTableSize();
    if(g_dot_.getTableSize()>0) size = std::min(size,g_dot_.getTableSize());
    if(g_nlo_.getTableSize()>0) size = std::min(size,g_nlo_.getTableSize());
    return size;
  }

  HellTableP & HellTableP::setMellinFunction(MellinFunction mf) {
    Real n = getTableSize();
    MFvals_.resize(n+1);
    //
    for (int i=0; i<n+1; ++i) {
      Complex N = getN(i);
      Complex g     = getGamma(i);
      Complex g_dot = getDotGamma(i);
      Complex g_nlo = getNLOGamma(i);
      MFvals_[i] = (*mf)(g, g_dot, N, getAlphaS(), getNF(), g_nlo);
    }
    return *this;
  }


  Real HellTableP::evalSimpson(Real z)
  {
    Real z_max = inv_mellin_max_z(z);
    Real u_max = z_max/(1+z_max);
    const Real TABLE_U_MAX = g_.step_*(getTableSize()-1);
    if(u_max > TABLE_U_MAX) u_max = TABLE_U_MAX;
    Real n = 4*static_cast<int>(u_max/g_.step_*0.25) + 4;
    //Real n = 6*static_cast<int>(z_max/g_.step_/6) + 6;
    u_max = n*g_.step_;
    if(u_max > TABLE_U_MAX) n -= 4;
    u_max = n*g_.step_;
    if(u_max > TABLE_U_MAX) abort(); // it should never happen
    //
    std::vector<Real> fvals(n+1);
    for (int i=0; i<n+1; ++i) {
      fvals[i] = imag(g_.eith_*pow(z,-getN(i)) * MFvals_[i]) * getJac(i);
    }
    Real * err_ptr = use_err_ ? &err_ : 0;

    return SimpsonIntegration::integrate(&fvals[0],0.0,u_max,n,err_ptr)/PI;
  }

  Real HellTableP::GaussIntegrand(Real t) {
    const Real TABLE_U_MAX = g_.step_*(getTableSize()-1);
    Real u = t/(1+t);
    if(u>TABLE_U_MAX) return 0;
    double ii = u/g_.step_;
    int i = (int)ii; // i <= ii
    //std::cout << i << "  " << ii << std::endl;
    //Real u = i*g_.step_;
    Complex nt = g_.c_ + t*g_.eith_ -1.;
    Real factor = ii-i;
    Complex gg = (1-factor)*MFvals_[i] + factor*MFvals_[i+1]; // interpolation
    Real res = imag(g_.eith_*pow(x_,-nt) * gg);
    //std::cout << res << std::endl;
    return res;
  }
  double integrand(double x, void *p) {
    HellTableP *htp = (HellTableP *) p;
    return htp->GaussIntegrand(x);
  }
  Real HellTableP::evalGauss(Real z, int &status)
  {
    x_ = z;
    double prec = 1e-2, error;
    return integration_gauss(integrand, 0., 33./log(1./z), prec, &error, this, &status)/PI;
  }


  Real HellTableP::getIntegrand(Real z, int i) {
    Real dt = getJac(i);
    Complex nt = getN(i);
    return imag(g_.eith_*pow(z,-nt) * MFvals_[i])*dt;
  }
  Real HellTableP::getIntegrand(MellinFunction mf, Real z, int i) {
    Real dt = getJac(i);
    Complex nt = getN(i);
    Complex g = getGamma(i);
    Complex g_dot = getDotGamma(i);
    Complex g_nlo = getNLOGamma(i);
    if (mf)
      g = (*mf)(g, g_dot, nt, getAlphaS(), getNF(), g_nlo);
    //
    return imag(g_.eith_*pow(z,-nt) * g)*dt;
  }

  Real HellTableP::eval(Real z) {
    return evalSimpson(z);
    Real result = 0;
    // I'd like to compute the integral with gauss (faster) and, if there is an error, ignore it and use Simpson (slower).
    gsl_error_handler_t *old_handler = gsl_set_error_handler (NULL);
    gsl_set_error_handler_off();
    int status;
    result = evalGauss(z, status);
    gsl_set_error_handler (old_handler);
    if(status) {
      std::cout << z << std::endl;
      result = evalSimpson(z);
    }
    return result;
  }

  Real HellTableP::getJac(unsigned i) const {
    Real u = i*g_.step_;
    Real i1mu = 1/(1-u);
    return i1mu*i1mu;
  }
  Complex HellTableP::getNPoles_(GammaHellTable g, Complex nt) const {
    Complex pp = 0;
    for (Poles p = g.poles_.begin(); p != g.poles_.end(); ++p)
      pp += (*p).residue/(nt - (*p).pole);
    return pp;
  }

  
} // namespace hell

